@@ -21,7 +21,9 @@ from mozphab.helpers import (
 from mozphab.logger import logger
 from mozphab.spinner import wait_message
 from mozphab.subprocess_wrapper import check_call_by_line
+from mozphab.telemetry import telemetry
 
+# The DEFAULT_UPDATE_MESSAGE is only required when using arc to submit.
 DEFAULT_UPDATE_MESSAGE = "Revision updated."
 
 REVISION_URL_RE = re.compile(r"^\s*Revision URI: (http.+)$", flags=re.MULTILINE)
@@ -107,7 +109,12 @@ def amend_revision_url(body, new_url):
 
 
 def show_commit_stack(
-    commits, validate=True, ignore_reviewers=False, show_rev_urls=False
+    commits,
+    wip=None,
+    validate=True,
+    ignore_reviewers=False,
+    show_rev_urls=False,
+    show_updated_only=False,
 ):
     """Log the commit stack in a human readable form."""
 
@@ -122,12 +129,23 @@ def show_commit_stack(
         ids = [int(c["rev-id"]) for c in commits if c.get("rev-id")]
         if ids:
             with wait_message("Loading existing revisions..."):
-                conduit.get_revisions(ids=ids)
+                revisions = conduit.get_revisions(ids=ids)
+
+            # preload diffs
+            with wait_message("Loading diffs..."):
+                diffs = conduit.get_diffs([r["fields"]["diffPHID"] for r in revisions])
 
     for commit in reversed(commits):
+        if show_updated_only and not commit["submit"]:
+            continue
+
+        closed = False
         change_bug_id = False
         is_author = True
         revision = None
+        wip_removed = False
+        wip_set = False
+        reviewers_added = False
 
         if commit.get("rev-id"):
             action = action_template % ("D" + commit["rev-id"])
@@ -143,47 +161,108 @@ def show_commit_stack(
                         and (commit["bug-id"] != revision["fields"]["bugzilla.bug-id"])
                     )
 
+                    # Check if revision is closed
+                    closed = revision["fields"]["status"]["closed"]
+
                     # Check if comandeering is required
                     whoami = conduit.whoami()
                     if "authorPHID" in revision["fields"] and (
                         revision["fields"]["authorPHID"] != whoami["phid"]
                     ):
                         is_author = False
+
+                    # Do we remove "Changes Planned" status?
+                    wip_removed = (
+                        not wip
+                        and revision["fields"]["status"]["value"] == "changes-planned"
+                    )
+
+                    # Do we set "Changes Planned" status?
+                    wip_set = (
+                        wip
+                        and revision["fields"]["status"]["value"] != "changes-planned"
+                    )
+
+                    # Any reviewers added to a revision without them?
+                    reviewers_added = bool(
+                        not revision["attachments"]["reviewers"]["reviewers"]
+                        and commit["reviewers"]["granted"]
+                    )
+
+                    # if SHA1 hasn't changed
+                    # and we're not changing the WIP status
+                    # and we're not adding reviewers to a revision without reviewers
+                    # and we're not changing the bug-id
+                    sha1_changed = (
+                        commit["node"]
+                        != diffs[revision["fields"]["diffPHID"]]["attachments"][
+                            "commits"
+                        ]["commits"][0]["identifier"]
+                    )
+                    if (
+                        not sha1_changed
+                        and not wip_removed
+                        and not wip_set
+                        and not reviewers_added
+                        and not change_bug_id
+                        and not closed
+                    ):
+                        commit["submit"] = False
+
         else:
             action = action_template % "New"
 
         logger.info("%s %s %s", action, commit["name"], commit["title-preview"])
         if validate:
-            if change_bug_id:
-                logger.warning(
-                    "!! Bug ID in Phabricator revision will be changed from %s to %s",
-                    revision["fields"]["bugzilla.bug-id"],
-                    commit["bug-id"],
+            if not commit["submit"]:
+                logger.info(
+                    " * This revision is not changed and will not be submitted."
                 )
 
-            if not is_author:
-                logger.warning(
-                    "!! You don't own this revision. Normally, you should only\n"
-                    '   update revisions you own. You can "Commandeer" this\n'
-                    "   revision from the web interface if you want to become\n"
-                    "   the owner."
-                )
+            else:
+                if wip_removed:
+                    logger.warning(
+                        '!! "Changes Planned" status will change to "Request Review"'
+                    )
 
-            if not commit["bug-id"]:
-                logger.warning("!! Missing Bug ID")
+                if change_bug_id:
+                    logger.warning(
+                        "!! Bug ID in Phabricator revision will change from %s to %s",
+                        revision["fields"]["bugzilla.bug-id"],
+                        commit["bug-id"],
+                    )
 
-            if commit["bug-id-orig"] and commit["bug-id"] != commit["bug-id-orig"]:
-                logger.warning(
-                    "!! Bug ID changed from %s to %s",
-                    commit["bug-id-orig"],
-                    commit["bug-id"],
-                )
+                if not is_author:
+                    logger.warning(
+                        "!! You don't own this revision. Normally, you should only\n"
+                        '   update revisions you own. You can "Commandeer" this\n'
+                        "   revision from the web interface if you want to become\n"
+                        "   the owner."
+                    )
 
-            if (
-                not ignore_reviewers
-                and not commit["reviewers"]["granted"] + commit["reviewers"]["request"]
-            ):
-                logger.warning("!! Missing reviewers")
+                if closed:
+                    logger.warning(
+                        "!! This revision is closed!\n"
+                        "   It will be reopened if submission proceeds.\n"
+                        "   You can stop now and refine the stack range."
+                    )
+
+                if not commit["bug-id"]:
+                    logger.warning("!! Missing Bug ID")
+
+                if commit["bug-id-orig"] and commit["bug-id"] != commit["bug-id-orig"]:
+                    logger.warning(
+                        "!! Bug ID changed from %s to %s",
+                        commit["bug-id-orig"],
+                        commit["bug-id"],
+                    )
+
+                if (
+                    not ignore_reviewers
+                    and not commit["reviewers"]["granted"]
+                    + commit["reviewers"]["request"]
+                ):
+                    logger.warning("!! Missing reviewers")
 
         if show_rev_urls and commit["rev-id"]:
             logger.warning("-> %s/D%s", conduit.repo.phab_url, commit["rev-id"])
@@ -326,6 +405,7 @@ def submit(repo, args):
     if environment.DEBUG:
         arcanist.ARC.append("--trace")
 
+    telemetry.metrics.mozphab.submission.preparation_time.start()
     with wait_message("Checking connection to Phabricator."):
         # Check if raw Conduit API can be used
         if not conduit.check():
@@ -342,7 +422,7 @@ def submit(repo, args):
 
     # Find and preview commits to submits.
     with wait_message("Looking for commits.."):
-        commits = repo.commit_stack()
+        commits = repo.commit_stack(single=args.single)
     if not commits:
         raise Error("Failed to find any commits to submit")
     logger.warning(
@@ -359,11 +439,13 @@ def submit(repo, args):
         update_commits_from_args(commits, args)
 
     # Validate commit stack is suitable for review.
-    show_commit_stack(commits, validate=True, ignore_reviewers=args.wip)
+    show_commit_stack(commits, wip=args.wip, validate=True, ignore_reviewers=args.wip)
     try:
         with wait_message("Checking commits.."):
             repo.check_commits_for_submit(
-                commits, validate_reviewers=not args.wip, require_bug=not args.no_bug,
+                commits,
+                validate_reviewers=not args.wip,
+                require_bug=not args.no_bug,
             )
     except Error as e:
         if not args.force:
@@ -390,6 +472,9 @@ def submit(repo, args):
             "be result in a comment on new revisions."
         )
 
+    telemetry.metrics.mozphab.submission.preparation_time.stop()
+    telemetry.metrics.mozphab.submission.commits_count.add(len(commits))
+
     # Confirmation prompt.
     if args.yes:
         pass
@@ -409,6 +494,7 @@ def submit(repo, args):
             config.write()
 
     # Process.
+    telemetry.metrics.mozphab.submission.process_time.start()
     previous_commit = None
     # Collect all existing revisions to get reviewers info.
     rev_ids_to_update = [int(c["rev-id"]) for c in commits if c.get("rev-id")]
@@ -419,8 +505,16 @@ def submit(repo, args):
 
         revisions_to_update = {str(r["id"]): r for r in list_to_update}
 
+    last_node = commits[-1]["orig-node"]
     for commit in commits:
+        diff = None
+
+        check_in_needed = args.check_in_needed and commit["orig-node"] == last_node
         # Only revisions being updated have an ID.  Newly created ones don't.
+        if not commit["submit"]:
+            previous_commit = commit
+            continue
+
         is_update = bool(commit["rev-id"])
         revision_to_update = (
             revisions_to_update[commit["rev-id"]] if is_update else None
@@ -471,13 +565,12 @@ def submit(repo, args):
                 diff = repo.get_diff(commit)
 
             if diff:
+                telemetry.metrics.mozphab.submission.files_count.add(len(diff.changes))
                 with wait_message("Uploading binary file(s)..."):
                     diff.upload_files()
 
                 with wait_message("Submitting the diff..."):
-                    diff_phid = diff.submit(commit, message)
-            else:
-                diff_phid = None
+                    diff.submit(commit, message)
 
             if is_update:
                 with wait_message("Updating revision..."):
@@ -485,9 +578,10 @@ def submit(repo, args):
                         commit,
                         has_commit_reviewers,
                         existing_reviewers,
-                        diff_phid=diff_phid,
+                        diff_phid=diff.phid,
                         wip=args.wip,
                         comment=args.message,
+                        check_in_needed=check_in_needed,
                     )
             else:
                 with wait_message("Creating a new revision..."):
@@ -495,9 +589,10 @@ def submit(repo, args):
                         commit,
                         commit["title-preview"],
                         summary,
-                        diff_phid,
+                        diff.phid,
                         has_commit_reviewers,
                         wip=args.wip,
+                        check_in_needed=check_in_needed,
                     )
 
             revision_url = "%s/D%s" % (repo.phab_url, rev["object"]["id"])
@@ -571,6 +666,11 @@ def submit(repo, args):
             with wait_message("Updating commit.."):
                 repo.amend_commit(commit, commits)
 
+        # Diff property has to be set after potential SHA1 change.
+        if args.no_arc and diff:
+            with wait_message("Setting diff metadata..."):
+                diff.set_property(commit, message)
+
         previous_commit = commit
 
     # Cleanup (eg. strip nodes) and refresh to ensure the stack is right for the
@@ -582,14 +682,17 @@ def submit(repo, args):
         repo.refresh_commit_stack(commits)
 
     logger.warning("\nCompleted")
-    show_commit_stack(commits, validate=False, show_rev_urls=True)
+    show_commit_stack(
+        commits, validate=False, show_rev_urls=True, show_updated_only=True
+    )
+    telemetry.metrics.mozphab.submission.process_time.stop()
 
 
 def add_parser(parser):
     submit_parser = parser.add_parser(
         "submit",
         aliases=["upload"],
-        help="Submit commit(s) to Phabricator.",
+        help="Submit commit(s) to Phabricator",
         description=(
             "MozPhab will change the working directory and amend the commits during "
             "the submission process."
@@ -613,7 +716,7 @@ def add_parser(parser):
     submit_parser.add_argument(
         "--message",
         "-m",
-        help="Provide a custom update message (default: %s)" % DEFAULT_UPDATE_MESSAGE,
+        help="Provide a custom update message (default: none)",
     )
     submit_parser.add_argument(
         "--force",
@@ -624,8 +727,8 @@ def add_parser(parser):
     submit_parser.add_argument(
         "--force-delete",
         action="store_true",
-        help="Mercurial only. Override the fail if a DAG branch point detected "
-        "and no evolve installed",
+        help="Mercurial only: Ignore error caused by a DAG branch point without "
+        "evolve installed",
     )
     submit_parser.add_argument(
         "--bug", "-b", help="Set Bug ID for all commits (default: from commit)"
@@ -633,7 +736,7 @@ def add_parser(parser):
     submit_parser.add_argument(
         "--no-bug",
         action="store_true",
-        help="Continue if a bug number is not provided.",
+        help="Continue if a bug number is not provided",
     )
     submit_parser.add_argument(
         "--reviewer",
@@ -655,6 +758,11 @@ def add_parser(parser):
         action="store_true",
         help="Do not run lint (default: lint changed files if configured)",
     )
+    submit_parser.add_argument(
+        "--check-in-needed",
+        action="store_true",
+        help="Add a `check-in-needed tag to the top most revision",
+    )
     submit_parser.add_argument(
         "--wip",
         "--plan-changes",
@@ -684,24 +792,30 @@ def add_parser(parser):
         "--remote",
         "-u",
         action="append",
-        help='Set upstream branch to detect the starting commit. (default: "")',
+        help='Set upstream branch to detect the starting commit (default: "")',
     )
     submit_parser.add_argument(
-        "--arc", dest="no_arc", action="store_false", help="Submits with Arcanist.",
+        "--arc",
+        dest="no_arc",
+        action="store_false",
+        help="Submits with Arcanist",
     )
     submit_parser.add_argument(
         "--force-vcs",
         action="store_true",
-        help="EXPERIMENTAL: Override VCS compatibility check.",
+        help="EXPERIMENTAL: Override VCS compatibility check",
     )
     submit_parser.add_argument(
         "--safe-mode",
         dest="safe_mode",
         action="store_true",
-        help="Run VCS with only necessary extensions.",
+        help="Run VCS with only necessary extensions",
     )
     submit_parser.add_argument(
-        "--single", "-s", action="store_true", help="Submit a single commit.",
+        "--single",
+        "-s",
+        action="store_true",
+        help="Submit a single commit",
     )
     submit_parser.add_argument(
         "start_rev",
