@@ -9,6 +9,8 @@ import subprocess
 import sys
 import uuid
 
+from datetime import datetime
+
 from mozphab import environment
 
 from .config import config
@@ -19,6 +21,7 @@ from .helpers import prompt, short_node, temporary_binary_file, temporary_file
 from .logger import logger
 from .repository import Repository
 from .spinner import wait_message
+from .telemetry import telemetry
 
 NULL_SHA1 = "0" * 40
 
@@ -42,6 +45,11 @@ class Git(Repository):
         super().__init__(path, dot_path)
 
         self.vcs = "git"
+        m = re.search(r"\d+\.\d+\.\d+", self.git.output(["--version"], split=False))
+        if not m:
+            raise Error("Failed to determine Git version.")
+
+        self.vcs_version = m.group(0)
         self.revset = None
         self.branch = None
 
@@ -254,11 +262,16 @@ class Git(Repository):
         Returns: A list of "aaaa bbbb cccc"" strings, where bbbb and cccc are
             SHA1 of direct children of aaaa
         """
-        return self.git_out(["rev-list", "--all", "--children", "--not", "%s^@" % node])
+
+        # Logging is disabled for this command as it can generate a _lot_ of
+        # results, especially on mozilla-central.
+        return self.git_out(
+            ["rev-list", "--all", "--children", "--not", "%s^@" % node], never_log=True
+        )
 
     @staticmethod
     def _get_direct_children(node, rev_list):
-        """ Return direct children of the commit.
+        """Return direct children of the commit.
 
         Args:
             node: The SHA1 of a node to check for direct children
@@ -329,7 +342,7 @@ class Git(Repository):
 
         return False
 
-    def commit_stack(self):
+    def commit_stack(self, single=False):
         """Collect all the info about commits."""
         if not self.revset:
             # No commits found to submit
@@ -353,15 +366,16 @@ class Git(Repository):
             ) = log_line.split("\n", 6)
             desc = desc.splitlines()
 
-            # Check if the commit is a child of the first one
-            if not rev_list:
-                rev_list = self._git_get_children(node)
-                first_node = node
-            elif not self._is_child(first_node, node, rev_list):
-                raise Error(
-                    "Commit %s is not a child of %s, unable to continue"
-                    % (short_node(node), short_node(first_node))
-                )
+            if not single:
+                # Check if the commit is a child of the first one
+                if rev_list is None:
+                    rev_list = self._git_get_children(node)
+                    first_node = node
+                elif not self._is_child(first_node, node, rev_list):
+                    raise Error(
+                        "Commit %s is not a child of %s, unable to continue"
+                        % (short_node(node), short_node(first_node))
+                    )
 
             # Check if commit has multiple parents, if so - raise an Error
             # We may push the merging commit if it's the first one
@@ -372,11 +386,16 @@ class Git(Repository):
                     % short_node(node)
                 )
 
+            # Tue, 14 Apr 2020 12:02:20 +0000
+            commit_epoch = datetime.strptime(
+                author_date, "%a, %d %b %Y %H:%M:%S %z"
+            ).timestamp()
             commits.append(
                 {
                     "name": short_node(node),
                     "node": node,
                     "orig-node": node,
+                    "submit": True,
                     "title": desc[0],
                     "title-preview": desc[0],
                     "body": "\n".join(desc[1:]).rstrip(),
@@ -386,6 +405,7 @@ class Git(Repository):
                     "parent": parents[0],
                     "tree-hash": tree_hash,
                     "author-date": author_date,
+                    "author-date-epoch": commit_epoch,
                     "author-name": author_name,
                     "author-email": author_email,
                 }
@@ -499,6 +519,9 @@ class Git(Repository):
 
         self.commit(body, author, author_date)
 
+    def format_patch(self, diff, body, author, author_date):
+        return diff
+
     def _get_current_head(self):
         """Return current's HEAD symbolic link."""
         symbolic = self.git_out(["symbolic-ref", "HEAD"], split=False)
@@ -635,6 +658,7 @@ class Git(Repository):
             b_size = self._file_size(b_blob)
 
         file_size = max(a_size, b_size)
+        telemetry.metrics.mozphab.submission.files_size.accumulate(file_size)
 
         # Detect if we're binary, and generate a unified diff
         if b"\0" in a_body or b"\0" in b_body or file_size > environment.MAX_TEXT_SIZE:
@@ -653,48 +677,68 @@ class Git(Repository):
                 change.binary = True
 
         if change.binary:
-            a_mime = mimetypes.guess_type(a_path)[0] or ""
-            b_mime = mimetypes.guess_type(b_path)[0] or ""
-            change.uploads = [
-                {"type": "old", "value": a_body, "mime": a_mime, "phid": None},
-                {"type": "new", "value": b_body, "mime": b_mime, "phid": None},
-            ]
-            if a_mime.startswith("image/") or b_mime.startswith("image/"):
-                change.file_type = Diff.FileType("IMAGE")
-            else:
-                change.file_type = Diff.FileType("BINARY")
+            change.set_as_binary(
+                a_body=a_body,
+                a_mime=mimetypes.guess_type(a_path)[0] or "",
+                b_body=b_body,
+                b_mime=mimetypes.guess_type(b_path)[0] or "",
+            )
+
         else:
             # We can only diff changed blobs.
             if a_blob == b_blob:
                 # No changes in the file contents.
-                lines = a_body.splitlines(True)
-                lines = [" %s" % l for l in lines]
-                old_off = new_off = 1
-                old_len = new_len = len(lines)
+                lines = a_body.splitlines(keepends=True)
+                if lines:
+                    lines = [" %s" % line for line in lines]
+                    change.hunks.append(
+                        Diff.Hunk(
+                            old_off=1,
+                            old_len=len(lines),
+                            new_off=1,
+                            new_len=len(lines),
+                            lines=lines,
+                        )
+                    )
+
             elif a_blob is None:
                 # The file is created.
-                lines = b_body.splitlines(True)
-                lines = ["+%s" % l for l in lines]
-                new_len = len(lines)
-                if lines and not lines[-1].endswith("\n"):
-                    lines[-1] = "{}\n".format(lines[-1])
-                    lines.append("\\ No newline at end of file\n")
+                lines = b_body.splitlines(keepends=True)
+                if lines:
+                    new_len = len(lines)
+                    lines = ["+%s" % line for line in lines]
+                    if lines and not lines[-1].endswith("\n"):
+                        lines[-1] = "{}\n".format(lines[-1])
+                        lines.append("\\ No newline at end of file\n")
+                    change.hunks.append(
+                        Diff.Hunk(
+                            old_off=0,
+                            old_len=0,
+                            new_off=1,
+                            new_len=new_len,
+                            lines=lines,
+                        )
+                    )
 
-                old_off = 0
-                new_off = 1
-                old_len = 0
             elif b_blob is None:
                 # The file is removed.
-                lines = a_body.splitlines(True)
-                lines = ["-%s" % l for l in lines]
-                old_len = len(lines)
-                if lines and not lines[-1].endswith("\n"):
-                    lines[-1] = "{}\n".format(lines[-1])
-                    lines.append("\\ No newline at end of file\n")
+                lines = a_body.splitlines(keepends=True)
+                if lines:
+                    old_len = len(lines)
+                    lines = ["-%s" % line for line in lines]
+                    if lines and not lines[-1].endswith("\n"):
+                        lines[-1] = "{}\n".format(lines[-1])
+                        lines.append("\\ No newline at end of file\n")
+                    change.hunks.append(
+                        Diff.Hunk(
+                            old_off=1,
+                            old_len=old_len,
+                            new_off=0,
+                            new_len=0,
+                            lines=lines,
+                        )
+                    )
 
-                old_off = 1
-                new_off = 0
-                new_len = 0
             else:
                 # There are changes in the file.
                 if self.args.lesscontext or file_size > environment.MAX_CONTEXT_SIZE:
@@ -712,39 +756,8 @@ class Git(Repository):
                     a_blob,
                     b_blob,
                 ]
-                git_diff = self.git_out(diff_args, expect_binary=True)
-                git_diff = str(git_diff, "utf-8").splitlines(keepends=True)
-                lines = git_diff[4:]
-                old_off, new_off, old_len, new_len = Diff.parse_git_diff(lines.pop(0))
-
-            # Collect some stats about the diff, and generate the corpus we
-            # want to send to Phabricator.
-            old_eof_newline = True
-            new_eof_newline = True
-            old_line = " "
-            corpus = "".join(lines)
-            for line in lines:
-                if line.endswith("No newline at end of file\n"):
-                    if old_line[0] != "+":
-                        old_eof_newline = False
-                    if old_line[0] != "-":
-                        new_eof_newline = False
-                old_line = line
-
-            change.hunks = [
-                diff.Hunk(
-                    old_off=old_off,
-                    old_len=old_len,
-                    new_off=new_off,
-                    new_len=new_len,
-                    old_eof_newline=old_eof_newline,
-                    new_eof_newline=new_eof_newline,
-                    added=sum(1 for l in lines if l[0] == "+"),
-                    deleted=sum(1 for l in lines if l[0] == "-"),
-                    corpus=corpus,
-                )
-            ]
-            change.file_type = diff.FileType("TEXT")
+                git_diff = self.git_out(diff_args, expect_binary=True).decode("utf-8")
+                change.from_git_diff(git_diff)
 
         diff.set_change_kind(change, kind_l[0], a_mode, b_mode, a_path, b_path)
 
@@ -773,13 +786,12 @@ class Git(Repository):
         return diff
 
     def check_vcs(self):
-        try:
-            return super().check_vcs()
-        except Error:
-            if not self.is_cinnabar_installed:
-                logger.error(
-                    "Git Cinnabar extension is required to work on this repository."
-                )
-                raise
+        if self.args.force_vcs:
+            return True
+
+        if self.is_cinnabar_required and not self.is_cinnabar_installed:
+            raise Error(
+                "Git Cinnabar extension is required to work on this repository."
+            )
 
         return True
