@@ -10,7 +10,9 @@ import uuid
 
 from contextlib import suppress
 from distutils.version import LooseVersion
+from functools import lru_cache
 
+import hglib
 from mozphab import environment
 
 from .config import config
@@ -26,7 +28,8 @@ from .helpers import (
 from .logger import logger
 from .repository import Repository
 from .spinner import wait_message
-from .subprocess_wrapper import check_call, check_output
+from .subprocess_wrapper import debug_log_command
+from .telemetry import telemetry
 
 MINIMUM_MERCURIAL_VERSION = LooseVersion("4.3.3")
 
@@ -40,8 +43,7 @@ class Mercurial(Repository):
 
         super().__init__(path, dot_path)
         self.vcs = "hg"
-
-        self._hg = config.hg_command.copy()
+        self._hg_binary = config.hg_command[0]
         self.revset = None
         self.strip_nodes = []
         self.status = None
@@ -53,26 +55,60 @@ class Mercurial(Repository):
         self.previous_bookmark = None
         self.has_temporary_bookmark = False
 
-        # Normalise/standardise Mercurial's output.
-        os.environ["HGPLAIN"] = "1"
-        os.environ["HGENCODING"] = "UTF-8"
+        # Check for `hg` presence
+        if not which_path(self._hg_binary):
+            raise Error(f"Failed to find hg executable ({self._hg_binary})")
 
-        # Check for `hg`, and mercurial version.
-        if not which_path(self._hg[0]):
-            raise Error("Failed to find hg executable ({})".format(self._hg[0]))
-        m = re.search(
-            r"\(version ([^)]+)\)", self.hg_out(["--version", "--quiet"], split=False)
-        )
-        if not m:
-            raise Error("Failed to determine Mercurial version.")
-        self.mercurial_version = LooseVersion(m.group(1))
+        self._config_options = {}
+        self._safe_config_options = {}
+        self._extra_options = {}
+        self._safe_mode = False
+        self._repo_path = path
+        hglib.HGPATH = self._hg_binary
+        self._repo = None
+        self._configs = []
+        major, minor, micro, *_ = self.repository.version
+        self.mercurial_version = LooseVersion(f"{major}.{minor}.{micro}")
+        self.vcs_version = str(self.mercurial_version)
         if self.mercurial_version < MINIMUM_MERCURIAL_VERSION:
             raise Error(
-                "You are currently running Mercurial %s.  "
-                "Mercurial %s or newer is required."
-                % (self.mercurial_version, MINIMUM_MERCURIAL_VERSION)
+                f"You are currently running Mercurial {self.mercurial_version}.  "
+                f"Mercurial {MINIMUM_MERCURIAL_VERSION} or newer is required."
             )
 
+        # Some Mercurial commands require the current working directory to be
+        # set to the repository root (eg. `hg files`).
+        os.chdir(self._repo_path)
+
+    @property
+    def repository(self):
+        """Returns the hglib.hgclient instance.
+
+        If the config has changed, recreate the instance.
+        """
+        configs = [f"{key}={value}" for key, value in self._get_config_options()]
+        configs.sort()
+        # returns the repo instance if the config has not changed.
+        if self._repo is not None and self._configs == configs:
+            return self._repo
+
+        if self._repo:
+            self._repo.close()
+
+        self._configs = configs
+        self._repo = hglib.open(self._repo_path, encoding="UTF-8", configs=configs)
+        return self._repo
+
+    def _get_config_options(self):
+        """Returns the --config options for hg
+
+        Updated with the safe ones when safe mode is used.
+        """
+        options = dict(self._config_options)
+        if self._safe_mode:
+            options.update(self._safe_config_options)
+        return list(options.items())
+
     @classmethod
     def is_repo(cls, path):
         """Quick check for repository at specified path."""
@@ -114,10 +150,45 @@ class Mercurial(Repository):
         return not status["T"]
 
     def hg(self, command, **kwargs):
-        check_call(self._hg + command, cwd=self.path, **kwargs)
+        self.hg_out(command, **kwargs)
 
-    def hg_out(self, command, **kwargs):
-        return check_output(self._hg + command, cwd=self.path, **kwargs)
+    def hg_out(
+        self,
+        command,
+        expect_binary=False,
+        strip=True,
+        keep_ends=False,
+        split=True,
+        never_log=False,
+    ):
+        def error_handler(exit_code, stdout, stderr):
+            if stdout and not never_log:
+                logger.debug(stdout.decode())
+
+            if stderr and not never_log:
+                logger.debug(stderr.decode())
+
+            raise CommandError(
+                "command '%s' failed to complete successfully" % command[0], exit_code
+            )
+
+        for arg, value in self._extra_options.items():
+            command.extend([arg, value])
+
+        debug_log_command(["hg"] + command)
+        command = [c.encode() for c in command]
+        out = self.repository.rawcommand(command, eh=error_handler)
+
+        if expect_binary:
+            logger.debug("%s bytes of data received", len(out))
+            return out
+
+        out = out.decode()
+        if strip:
+            out = out.rstrip()
+        if out and not never_log:
+            logger.debug(out)
+        return out.splitlines(keep_ends) if split else out
 
     def hg_log(self, revset, split=True, select="node"):
         return self.hg_out(["log", "-T", "{%s}\n" % select, "-r", revset], split=split)
@@ -128,8 +199,8 @@ class Mercurial(Repository):
         # be deleted in cleanup(). Mercurial will automatically move the
         # bookmark to the successors as we update commits.
         for active, bookmark in [
-            l.split(" ", 1)
-            for l in self.hg_out(["bookmark", "-T", "{active} {bookmark}\n"])
+            line.split(" ", 1)
+            for line in self.hg_out(["bookmark", "-T", "{active} {bookmark}\n"])
         ]:
             if active == "True":
                 self.previous_bookmark = bookmark
@@ -220,6 +291,13 @@ class Mercurial(Repository):
         super().refresh_commit_stack(commits)
 
     def set_args(self, args):
+        """Sets up the right environment for hg, prior to running it.
+
+        Sets:
+        - all the --config options (includes safe mode)
+        - extra options
+        - the `revset` attribute
+        """
         super().set_args(args)
 
         # Load hg config into hg_config.  We'll specify specific settings on
@@ -232,39 +310,38 @@ class Mercurial(Repository):
             ),
         )
 
-        safe_mode = self.args.safe_mode or config.safe_mode
-        safe_options = []
-        options = []
+        self._safe_mode = self.args.safe_mode or config.safe_mode
+        self._safe_config_options.clear()
+        self._config_options.clear()
+        self._extra_options.clear()
 
         # Need to use the correct username.
         if "ui.username" not in hg_config:
             raise Error("ui.username is not configured in your hgrc")
+        self._safe_config_options["ui.username"] = hg_config["ui.username"]
 
-        safe_options.extend(["--config", "ui.username=%s" % hg_config["ui.username"]])
+        # Always need rebase.
+        self._config_options["extensions.rebase"] = ""
 
-        options.extend(
-            # Always need rebase.
-            ["--config", "extensions.rebase="]
-            # Mercurial should never paginate the response.
-            + ["--pager", "never"]
-        )
+        # Mercurial should never paginate the response.
+        self._extra_options["--pager"] = "never"
 
         # Perform rebases in-memory to improve performance (requires v4.5+).
-        if not safe_mode and self.mercurial_version >= LooseVersion("4.5"):
-            options.extend(["--config", "rebase.experimental.inmemory=true"])
+        if not self._safe_mode and self.mercurial_version >= LooseVersion("4.5"):
+            self._config_options["rebase.experimental.inmemory"] = "true"
 
         # Enable evolve if the user's currently using it.  evolve makes amending
         # commits with children trivial (amongst other things).
         ext_evolve = self._get_extension("evolve", hg_config)
         if ext_evolve is not None:
-            safe_options.extend(["--config", "extensions.evolve=%s" % ext_evolve])
+            self._safe_config_options["extensions.evolve"] = ext_evolve
             self.use_evolve = True
 
         # Otherwise just enable obsolescence markers, and when we're done remove
         # the obsstore we created.
         else:
-            options.extend(["--config", "experimental.evolution.createmarkers=true"])
-            options.extend(["--config", "extensions.strip="])
+            self._config_options["experimental.evolution.createmarkers"] = "true"
+            self._config_options["extensions.strip"] = ""
             self.use_evolve = False
             self.obsstore = os.path.join(self.path, ".hg", "store", "obsstore")
             self.unlink_obsstore = not os.path.exists(self.obsstore)
@@ -273,30 +350,26 @@ class Mercurial(Repository):
         ext_mq = self._get_extension("mq", hg_config)
         self.has_mq = ext_mq is not None
         if self.has_mq:
-            safe_options.extend(["--config", "extensions.mq=%s" % ext_mq])
+            self._safe_config_options["extensions.mq"] = ext_mq
 
         # `shelve` is useful for dealing with uncommitted changes; track if it's
         # currently enabled so we can tailor our error accordingly.
         self.has_shelve = self._get_extension("shelve", hg_config) is not None
 
         # Disable the user's hgrc file, to ensure we run without rogue extensions.
-        if safe_mode:
+        if self._safe_mode:
             os.environ["HGRCPATH"] = ""
-            options.extend(safe_options)
-
+            options = self._get_config_options()
             logger.debug(
                 "hg extensions (safe mode): %s",
                 ", ".join(self._get_extensions(from_args=options)),
             )
-
         else:
             logger.debug(
                 "hg extensions: %s",
                 ", ".join(self._get_extensions(from_config=hg_config)),
             )
 
-        self._hg.extend(options)
-
         if hasattr(self.args, "start_rev"):
             is_single = hasattr(self.args, "single") and self.args.single
             # Set the default start revision.
@@ -332,14 +405,14 @@ class Mercurial(Repository):
 
             self.revset = "%s::%s" % (short_node(start), short_node(end))
 
-    def commit_stack(self):
+    def commit_stack(self, **kwargs):
         # Grab all the info we need about the commits, using randomness as a delimiter.
         boundary = "--%s--\n" % uuid.uuid4().hex
         hg_log = self.hg_out(
             ["log"]
             + [
                 "-T",
-                "{rev}\n{node}\n{date|rfc822date}\n{author|person}\n{author|email}\n"
+                "{rev}\n{node}\n{date|hgdate}\n{author|person}\n{author|email}\n"
                 "{desc}%s" % boundary,
             ]
             + ["-r", self.revset],
@@ -368,6 +441,7 @@ class Mercurial(Repository):
                 {
                     "name": "%s:%s" % (rev, short_node(node)),
                     "node": node,
+                    "submit": True,
                     "public-node": node,
                     "orig-node": node,
                     "title": desc[0],
@@ -376,7 +450,7 @@ class Mercurial(Repository):
                     "bug-id": None,
                     "reviewers": dict(request=[], granted=[]),
                     "rev-id": None,
-                    "author-date": author_date,
+                    "author-date-epoch": int(author_date.split(" ")[0]),
                     "author-name": author_name,
                     "author-email": author_email,
                 }
@@ -448,17 +522,18 @@ class Mercurial(Repository):
                 logger.info("Bookmark set to %s", bookmark_name)
 
     def apply_patch(self, diff, body, author, author_date):
+        changeset_str = self.format_patch(diff, body, author, author_date)
+        with temporary_binary_file(changeset_str.encode("utf8")) as changeset_file:
+            self.hg(["import", changeset_file, "--quiet"])
+
+    def format_patch(self, diff, body, author, author_date):
         changeset = ["# HG changeset patch"]
         if author:
             changeset.append("# User {}".format(author))
-
         if author_date:
             changeset.append("# Date {} 0".format(author_date))
-
         changeset.extend([body, diff])
-        changeset_str = "\n".join(changeset)
-        with temporary_binary_file(changeset_str.encode("utf8")) as changeset_file:
-            self.hg(["import", changeset_file, "--quiet"])
+        return "\n".join(changeset)
 
     def _amend_commit_body(self, node, body):
         with temporary_file(body) as body_file:
@@ -651,7 +726,7 @@ class Mercurial(Repository):
         # using a -T template here doesn't output all files (eg. source files from
         # a copy operation are skipped), so we have to parse the default output
         modified_files = [
-            line[2:]  # strip leading status char and space
+            line[2:].replace("\\", "/")  # strip leading status char and space
             for line in self.hg_out(["status", "--change", commit["node"], "--copies"])
         ]
 
@@ -782,9 +857,11 @@ class Mercurial(Repository):
 
         return diff
 
+    @lru_cache(maxsize=None)
     def hg_cat(self, fn, node):
         return self.hg_out(["cat", "-r", node, fn], split=False, expect_binary=True)
 
+    @lru_cache(maxsize=None)
     def _file_size(self, fn, rev):
         """Get the file size of the file."""
         return int(
@@ -794,6 +871,7 @@ class Mercurial(Repository):
             )
         )
 
+    @lru_cache(maxsize=128)
     def _get_file_meta(self, fn, rev):
         """Collect information about the file."""
         binary = False
@@ -822,157 +900,120 @@ class Mercurial(Repository):
 
         return meta
 
-    def _change_add(self, change, fn, _, parent, node):
+    def _change_add(self, change, fn, _old_fn, _parent, node):
         """Create a change about adding a file to the commit."""
         meta = self._get_file_meta(fn, node)
+        telemetry.metrics.mozphab.submission.files_size.accumulate(meta["file_size"])
+
         if meta["binary"]:
-            self._change_set_binary(change, "", meta["bin_body"], "", meta["mime"])
-        else:
-            lines = meta["body"].splitlines(keepends=True)
-            lines = ["+%s" % l for l in lines]
-
-            if lines and not lines[-1].endswith("\n"):
-                lines[-1] = "{}\n".format(lines[-1])
-                lines.append("\\ No newline at end of file\n")
-
-            self._change_create_hunk(
-                change, fn, lines, meta["file_size"], parent, node, 0, 1, 0, len(lines)
+            change.set_as_binary(
+                a_body="",
+                a_mime="",
+                b_body=meta["bin_body"],
+                b_mime=meta["mime"],
             )
+            return
 
-    def _change_del(self, change, fn, _, parent, node):
+        lines = meta["body"].splitlines(keepends=True)
+        lines = ["+%s" % line for line in lines]
+        new_len = len(lines)
+
+        if lines and not lines[-1].endswith("\n"):
+            lines[-1] = "{}\n".format(lines[-1])
+            lines.append("\\ No newline at end of file\n")
+
+        change.hunks.append(
+            Diff.Hunk(
+                old_off=0,
+                new_off=1,
+                old_len=0,
+                new_len=new_len,
+                lines=lines,
+            )
+        )
+
+    def _change_del(self, change, fn, _old_fn, parent, _node):
         """Create a change about deleting a file from the commit."""
         meta = self._get_file_meta(fn, parent)
+        telemetry.metrics.mozphab.submission.files_size.accumulate(meta["file_size"])
+
         if meta["binary"]:
-            self._change_set_binary(change, meta["bin_body"], "", meta["mime"], "")
-        else:
-            lines = meta["body"].splitlines(keepends=True)
-            lines = ["-%s" % l for l in lines]
-
-            if lines and not lines[-1].endswith("\n"):
-                lines[-1] = "{}\n".format(lines[-1])
-                lines.append("\\ No newline at end of file\n")
-
-            self._change_create_hunk(
-                change, fn, lines, meta["file_size"], parent, node, 1, 0, len(lines), 0
+            change.set_as_binary(
+                a_body=meta["bin_body"],
+                a_mime=meta["mime"],
+                b_body="",
+                b_mime="",
             )
+            return
+
+        lines = meta["body"].splitlines(keepends=True)
+        lines = ["-%s" % line for line in lines]
+
+        if lines and not lines[-1].endswith("\n"):
+            lines[-1] = "{}\n".format(lines[-1])
+            lines.append("\\ No newline at end of file\n")
+
+        old_len = len(lines)
+
+        change.hunks.append(
+            Diff.Hunk(
+                old_off=1,
+                new_off=0,
+                old_len=old_len,
+                new_len=0,
+                lines=lines,
+            )
+        )
 
     def _change_mod(self, change, fn, old_fn, parent, node):
         """Create a change about modified file in the commit."""
         a_meta = self._get_file_meta(old_fn, parent)
         b_meta = self._get_file_meta(fn, node)
+        file_size = max(a_meta["file_size"], b_meta["file_size"])
+        telemetry.metrics.mozphab.submission.files_size.accumulate(file_size)
+
         if a_meta["binary"] or b_meta["binary"]:
-            self._change_set_binary(
-                change,
-                a_meta["bin_body"],
-                b_meta["bin_body"],
-                a_meta["mime"],
-                b_meta["mime"],
+            # Binary file.
+            change.set_as_binary(
+                a_body=a_meta["bin_body"],
+                a_mime=a_meta["mime"],
+                b_body=b_meta["bin_body"],
+                b_mime=b_meta["mime"],
             )
-        else:
-            file_size = max(a_meta["file_size"], b_meta["file_size"])
-            if a_meta["body"] == b_meta["body"]:
-                lines = a_meta["body"].splitlines(True)
-                lines = [" %s" % l for l in lines]
-                old_off = new_off = 1
-                old_len = new_len = len(lines)
-            else:
-                if self.args.lesscontext or file_size > environment.MAX_CONTEXT_SIZE:
-                    context_size = 100
-                else:
-                    context_size = environment.MAX_CONTEXT_SIZE
+            return
 
-                # Monitoring if parsing the diff is not slowing down the submit process.
-                start = time.process_time()
-                git_diff = self.hg_out(
-                    ["diff", "--git", "-U%s" % context_size, "-r", parent, fn],
-                    expect_binary=True,
+        if a_meta["body"] == b_meta["body"]:
+            # File contents unchanged.
+            lines = a_meta["body"].splitlines(keepends=True)
+            lines = [" %s" % line for line in lines]
+            change.hunks.append(
+                Diff.Hunk(
+                    old_off=1,
+                    new_off=1,
+                    old_len=len(lines),
+                    new_len=len(lines),
+                    lines=lines,
                 )
-                git_diff = str(git_diff, "utf-8").splitlines(keepends=True)
-                # Remove all info above the header
-                lines = []
-                found_hdr = False
-                for line in git_diff:
-                    if not found_hdr:
-                        found_hdr = line.startswith("@@")
-
-                    if found_hdr:
-                        lines.append(line)
-
-                if file_size > environment.MAX_CONTEXT_SIZE / 2:
-                    logger.debug(
-                        "Splitting the diff (size: %s) took %ss",
-                        file_size,
-                        time.process_time() - start,
-                    )
-
-                old_off, new_off, old_len, new_len = Diff.parse_git_diff(lines.pop(0))
-
-            self._change_create_hunk(
-                change,
-                fn,
-                lines,
-                file_size,
-                parent,
-                node,
-                old_off,
-                new_off,
-                old_len,
-                new_len,
             )
+            return
 
-    def _change_set_binary(self, change, a_body, b_body, a_mime, b_mime):
-        """Sets `Change` object as binary."""
-        change.binary = True
-        change.uploads = [
-            {"type": "old", "value": a_body, "mime": a_mime, "phid": None},
-            {"type": "new", "value": b_body, "mime": b_mime, "phid": None},
-        ]
-        if a_mime.startswith("image/") or b_mime.startswith("image/"):
-            change.file_type = Diff.FileType("IMAGE")
+        if self.args.lesscontext or file_size > environment.MAX_CONTEXT_SIZE:
+            context_size = 100
         else:
-            change.file_type = Diff.FileType("BINARY")
+            context_size = environment.MAX_CONTEXT_SIZE
 
-    def _change_create_hunk(
-        self,
-        change,
-        fn,
-        lines,
-        file_size,
-        parent,
-        node,
-        old_off,
-        new_off,
-        old_len,
-        new_len,
-    ):
-        """Creates a hunk for the Change object.
+        start = time.process_time()
 
-        Collects some stats about the diff, and generates the corpus we
-        want to send to the Phabricator.
-        """
-        old_eof_newline = True
-        new_eof_newline = True
-        old_line = " "
-        corpus = "".join(lines)
-        for line in lines:
-            if line.endswith("No newline at end of file\n"):
-                if old_line[0] != "+":
-                    old_eof_newline = False
-                if old_line[0] != "-":
-                    new_eof_newline = False
-            old_line = line
+        # Using binary here to ensure we avoid converting newlines
+        git_diff = self.hg_out(
+            ["diff", "--git", "--unified", str(context_size), "--rev", parent, fn],
+            expect_binary=True,
+        ).decode("utf-8")
+        change.from_git_diff(git_diff)
 
-        change.hunks = [
-            Diff.Hunk(
-                old_off=old_off,
-                old_len=old_len,
-                new_off=new_off,
-                new_len=new_len,
-                old_eof_newline=old_eof_newline,
-                new_eof_newline=new_eof_newline,
-                added=sum(1 for l in lines if l[0] == "+"),
-                deleted=sum(1 for l in lines if l[0] == "-"),
-                corpus=corpus,
+        if file_size > environment.MAX_CONTEXT_SIZE / 2:
+            logger.debug(
+                "Splitting the diff (size: %s) took %ss",
+                file_size,
+                time.process_time() - start,
             )
-        ]
-        change.file_type = Diff.FileType("TEXT")
