@@ -9,6 +9,7 @@
 CLI to support submission of a series of commits to Phabricator. .
 """
 
+import logging
 import os
 import sys
 import traceback
@@ -24,6 +25,7 @@ from .exceptions import Error
 from .logger import init_logging, logger
 from .spinner import wait_message
 from .sentry import init_sentry, report_to_sentry
+from .telemetry import telemetry
 from .updater import check_for_updates, get_name_and_version
 
 # Known Issues
@@ -49,25 +51,34 @@ def main(argv, *, is_development):
 
         if args.trace:
             environment.DEBUG = True
+        if environment.DEBUG:
+            environment.SHOW_SPINNER = False
 
         init_logging()
         logger.debug(get_name_and_version())
 
-        with_arc = not hasattr(args, "no_arc") or not args.no_arc
-        if with_arc:
+        if not args.no_arc:
             install_arc_if_required()
 
-        if environment.DEBUG:
+        # Ensure that `patch --raw ..` only outputs the patch
+        if args.command == "patch" and getattr(args, "raw", False):
             environment.SHOW_SPINNER = False
+            logger.setLevel(logging.ERROR)
 
-        if args.command != "self-update":
-            check_for_updates(with_arc=with_arc)
+        elif args.command != "self-update":
+            check_for_updates(with_arc=not args.no_arc)
 
+        repo = None
         if args.needs_repo:
             with wait_message("Starting up.."):
                 repo = repo_from_args(args)
 
             conduit.set_repo(repo)
+
+        telemetry.set_metrics(args, is_development=is_development)
+
+        if repo is not None:
+            telemetry.set_vcs(repo)
             try:
                 args.func(repo, args)
             finally:
@@ -76,6 +87,9 @@ def main(argv, *, is_development):
         else:
             args.func(args)
 
+        telemetry.metrics.mozphab.usage.command_time.stop()
+        telemetry.submit()
+
     except KeyboardInterrupt:
         pass
     except Error as e:
