@@ -35,6 +35,7 @@ class Repository(object):
         self.phab_url = (phab_url or self._phab_url()).rstrip("/")
         self.api_url = self._api_url()
         self.call_sign = self._get_setting("repository.callsign")
+        self.bmo_url = self._get_setting("bmo_url")
 
     def is_worktree_clean(self):
         """Check if the working tree is clean."""
@@ -98,12 +99,13 @@ class Repository(object):
     def untracked(self):
         """Return a list of untracked files."""
 
-    def commit_stack(self):
+    def commit_stack(self, **kwargs):
         """Return list of commits.
 
         List of dicts with the following keys:
             name          human readable identifier of commit (eg. short sha)
-            node          sha/hash
+            node          SHA1 in stack
+            orig-node     an original SHA1 of the commit
             title         first line of commit description (unaltered)
             body          commit description, excluding first line
             title-preview title with bug-id and reviewer modifications
@@ -111,6 +113,11 @@ class Repository(object):
             bug-id-orig   original bug-id from commit desc
             reviewers     list of reviewers
             rev-id        phabricator revision id
+            parent        SHA1 of the parent commit
+            author-date   string representation of the commit creation time
+            author-date-epoch
+            author-name
+            author-email
         """
 
     def refresh_commit_stack(self, commits):
@@ -148,6 +155,9 @@ class Repository(object):
     def apply_patch(self, diff, body, author, author_date):
         """Apply the patch and commit the changes."""
 
+    def format_patch(self, diff, body, author, author_date):
+        """Format a patch appropriate for importing."""
+
     def check_commits_for_submit(
         self, commits, *, validate_reviewers=True, require_bug=True
     ):
@@ -278,8 +288,7 @@ class Repository(object):
         return False
 
     def _api_url(self):
-        """Return a base URL for conduit API call
-        """
+        """Return a base URL for conduit API call"""
         return urllib.parse.urljoin(self.phab_url, "api/")
 
     @property
